#!/usr/bin/env python3
# Copyright(c) 2020-2021 by craftyguy "Clayton Craft" <clayton@craftyguy.net>
# Distributed under GPLv3+ (see COPYING) WITHOUT ANY WARRANTY.
import argparse
import gi
import sys
from .window import MainWindow
from .logger import Logger
from .config import Config
from .preferences import PrefsWindow
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
gi.require_version('Notify', '0.7')
from gi.repository import Gtk, Gio, GLib
from gi.repository import Handy

APP_NAME = 'caerbannog'
APP_ID = f"net.craftyguy.{APP_NAME}"


class Application(Gtk.Application):
    """Gtk Application."""
    def __init__(self, version):
        super().__init__(application_id=APP_ID,
                         flags=Gio.ApplicationFlags.FLAGS_NONE)

        GLib.set_prgname(APP_ID)
        GLib.set_application_name(APP_NAME)
        self.window = None
        action = Gio.SimpleAction.new("quit", None)
        action.connect("activate", self.on_quit)
        self.add_action(action)
        self._config = Config()
        self._version = version
        self._settings = Gtk.Settings.get_default()
        self._settings.set_property("gtk-application-prefer-dark-theme",
                                    self._config.get('general', 'dark_theme',
                                                     False))

    def __add_action(self, action_name, callback):
        action = Gio.SimpleAction.new(action_name, None)
        action.connect("activate", callback)
        self.add_action(action)

    def do_startup(self):
        Gtk.Application.do_startup(self)
        Handy.init()

        # connect actions
        self.__add_action("quit", self.on_quit)
        self.__add_action("preferences", self.on_show_preferences)
        self.__add_action("about", self.on_show_about)

        self.window = MainWindow(title=APP_NAME, application=self,
                                 config=self._config)

    def do_activate(self):
        if not self.window:
            self.window = MainWindow(title=APP_NAME, application=self,
                                     config=self._config)
        self.window.present()

    def on_show_preferences(self, *args):
        prefs_window = PrefsWindow(config=self._config)
        prefs_window.connect("destroy", self.on_preferences_closed)
        prefs_window.set_transient_for(self.window)
        prefs_window.present()
        prefs_window.show_all()

    def on_preferences_closed(self, prefs_window):
        self.window.reload()

    def on_show_about(self, *args):
        about = Gtk.AboutDialog(transient_for=self.props.active_window,
                                copyright="© 2020-2021 Clayton Craft",
                                authors=["Clayton Craft"],
                                logo_icon_name=APP_ID,
                                license_type=Gtk.License.GPL_3_0,
                                version=str(self._version),
                                comments=("Frontend for password-store"),
                                website=f"https://sr.ht/~craftyguy/{APP_NAME}",
                                website_label=("Website"))
        about.run()
        about.destroy()

    def on_quit(self, *args):
        self.quit()


def main(version=None):
    parser = argparse.ArgumentParser(description='Frontend for password-store')
    parser.add_argument('-v', '--version', action='store_true',
                        help='Print version and exit')
    args = parser.parse_args()
    if args.version:
        Logger.info(f"version: {version}")
        sys.exit()
    app = Application(version)
    app.run()


if __name__ == "__main__":
    main()
