# Copyright(c) 2020-2021 by craftyguy "Clayton Craft" <clayton@craftyguy.net>
# Distributed under GPLv3+ (see COPYING) WITHOUT ANY WARRANTY.
import anytree
import fuzzyfinder
import gc
import gi
import importlib
import os
from .logger import Logger
gi.require_version('Gdk', '3.0')
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
gi.require_version('Notify', '0.7')
from gi.repository import Gdk, Gio, Gtk, GLib, GObject
from gi.repository import Notify

APP_NAME = 'caerbannog'
UI_PATH = f"/net/craftyguy/{APP_NAME}/ui/"


@Gtk.Template(resource_path=UI_PATH+'pass_store_list.ui')
class PassList(Gtk.ListBox):
    __gtype_name__ = 'PassList'
    __gsignals__ = {
        'pass-level-changed': (GObject.SignalFlags.RUN_LAST, None, (str, )),
    }

    def __init__(self, config):
        super().__init__()

        self._config = config
        self._pass_files = []
        self._pass_sections = {}
        self._pass_item_key_list = []
        self._filter_text = ''
        self._store_path = ''
        self.__no_files_row = None

        self.connect('row-selected', self.on_pass_item_selected)
        self._clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        Notify.init(APP_NAME)
        self._store = Gio.ListStore.new(Gtk.ListBoxRow)
        self._root_s = 'root'
        self._root = anytree.Node(self._root_s)
        self._tree_cur = self._root
        self.bind_model(self._store, self._get_widget_for_list)

    @property
    def current_tree_level(self):
        if self._tree_cur.name == self._root_s:
            return None
        return self._tree_cur.name

    def _get_widget_for_list(self, obj):
        return obj

    def __del__(self):
        Notify.uninit()

    def get_keys_from_store(self, force=False):
        if not self._pass_files or force:
            self._store_path = os.path.expanduser(self._config.get('general',
                                                                   'pass_store_home'))
            self.refresh_pass_files()
        ps_home = os.path.expanduser(self._config.get('general',
                                                      'pass_store_home'))
        # pass store home + /
        ps_len = len(ps_home) + 1
        for file in self._pass_files:
            # strip out the password-store path from the name of the pass item
            key = file[ps_len:]
            if key.endswith('.gpg'):
                key = key[:-4]
            self._pass_item_key_list.append(f"/{self._root_s}/{key}")
            self._import_keypath(key, file)

    def update_view(self, go_back=False):
        for row in self._store:
            row.hide()
        if go_back and not self._tree_cur.is_root:
            self._tree_cur = self._tree_cur.parent
        if not self._tree_cur.children:
            Logger.info("No password files found in {self._tree_cur}! Check "
                        "the path to the password-store in Preferences")
            if self._no_files_row not in self._store:
                self._store.append(self._no_files_row)
            self._no_files_row.show_all()
            return

        _, level = anytree.ZigZagGroupIter(self._tree_cur,
                                           maxlevel=2)
        for item in level:
            item.widget.show_all()
            if item.widget not in self._store:
                self._store.append(item.widget)

    @property
    def _no_files_row(self):
        if not self.__no_files_row:
            self.__no_files_row = Gtk.ListBoxRow()
            label = Gtk.Label()
            label.set_xalign(0.5)
            label.set_text("No password files found.")
            hbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
            hbox.pack_start(label, False, True, 0)
            self.__no_files_row.add(hbox)

            self.__no_files_row.set_margin_top(20)
            self.__no_files_row.set_selectable(False)
            self.__no_files_row.set_activatable(False)
        return self.__no_files_row

    def _import_keypath(self, keypath, file):
        path = keypath.split('/')
        path_len = len(path)
        parent = self._root
        resolver = anytree.Resolver('name')
        for i in range(path_len):
            is_leaf = i == path_len - 1
            try:
                parent = resolver.get(parent, path[i])
                # parent exists
                continue
            except anytree.resolver.ChildResolverError:
                if is_leaf:
                    PassItem(path[i], self._config, file=file, parent=parent)
                else:
                    # create new parent
                    parent = PassSection(path[i], parent=parent)

    def refresh_pass_files(self):
        self._pass_files = []
        # os.walk is a bit faster than glob.glob
        for p, _, f in os.walk(self._store_path):
            if '.git/' in p:
                continue
            for file in f:
                if file.endswith('.gpg'):
                    self._pass_files.append(os.path.join(self._root_s, p,
                                                         file))

    def update_filter(self, filter_text):
        """ Filter out items in current pass tree branch and add items to
        view """
        self._filter_text = filter_text
        if self._filter_text:
            filtered_keys = fuzzyfinder.fuzzyfinder(self._filter_text,
                                                    self._pass_item_key_list)
        else:
            # no text, so just display all items
            self.update_view()
            return
        resolver = anytree.Resolver('name')
        path = getattr(self._tree_cur, 'pathstr', '/')
        path = os.path.join('/', path, self._tree_cur.name)
        for row in self._store:
            row.hide()
        for key in filtered_keys:
            if not key.startswith(path):
                continue
            node = resolver.get(self._tree_cur, key)
            node.widget.show_all()
            if node.widget not in self._store:
                self._store.append(node.widget)

    def on_pass_item_selected(self, container, obj):
        if not obj:
            # signal fires when row selection is cleared too
            # https://developer.gnome.org/gtk3/stable/GtkListBox.html#GtkListBox-row-selected
            return
        node = obj.tree_node
        # order of this conditional is important, since PassItem is derived
        # from PassSection
        if isinstance(node, PassItem):
            self._copy_pass(node)
        elif isinstance(node, PassSection):
            # populate listbox with selected subsection
            self._tree_cur = node
            self.update_view()
            self.emit('pass-level-changed', node.name)

    def _copy_pass(self, pass_item):
        pw = pass_item.get_pass()
        if not pw:
            Logger.error("Unable to decrypt password, nothing copied to "
                         "clipboard!")
            notifier = Notify.Notification.new(APP_NAME,
                                               "Unable to copy password, "
                                               "decryption failed!")
            notifier.show()
            return
        self._clipboard.set_text(pw, -1)
        # schedule timer to clear clipboard
        del pw
        gc.collect()
        timeout = self._config.get('general', 'clipboard_timeout_seconds')
        try:
            timeout = int(timeout)
        except (ValueError, TypeError):
            Logger.error(f"Unable to parse timeout from: {timeout}")
            timeout = 45
            Logger.error(f"Using default timeout of {timeout} seconds")
        self.clear_cb_timer = GLib.timeout_add_seconds(timeout,
                                                       self.clear_clipboard)

        notifier = Notify.Notification.new(APP_NAME,
                                           "Password copied to clipboard!\n"
                                           "It will be cleared in "
                                           f"{timeout} seconds.")
        notifier.show()

    def clear_clipboard(self):
        """ Clears the clipboard """
        self._clipboard.set_text('', -1)
        Logger.info('Clipboard cleared!')

        notifier = Notify.Notification.new(APP_NAME,
                                           "Password cleared from clipboard!")
        notifier.show()
        self.unselect_all()


class PassSection(anytree.Node):

    """ Extends anytree.Node to provide methods for managing associated
    widgets, etc """
    pathstr: str = ''

    def __init__(self, name, parent=None, children=None, **kwargs):
        super().__init__(name, parent, children, **kwargs)

        self._row = None
        self.pathstr = '/'.join([str(node.name)
                                 for node in self.path
                                 if node.name != self.name])
        self.tree_node = None

    @property
    def widget(self):
        """ Return row rendering contents of node.
        Note: Reference to self is set in widget.tree_node """
        if self._row:
            return self._row
        row = Gtk.ListBoxRow()
        hbox = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        keylabel = Gtk.Label()
        keylabel.set_xalign(0.0)
        suffix = '/' if self.children else ''
        keylabel.set_markup(f"<b>{self.name}{suffix}</b>")

        path = self.pathstr[len('root/'):]
        # path label is hidden when at root
        if path:
            pathlabel = Gtk.Label()
            pathlabel.set_xalign(0.0)
            pathlabel.set_markup(f"<small>{path}</small>")
            # place label above the keylabel
            hbox.pack_start(pathlabel, False, True, 0)
        else:
            # center text when at the top level since there is no pathlabel
            # above
            keylabel.set_margin_top(5)
            keylabel.set_margin_bottom(5)

        hbox.pack_start(keylabel, False, True, 0)
        row.add(hbox)
        # store reference to self
        row.tree_node = self
        self._row = row

        return row


class PassItem(PassSection):
    """ PassSection tree node with methods for managing pass files """

    def __init__(self, name, config, file, parent=None, children=None,
                 **kwargs):
        super().__init__(name, parent, children, **kwargs)

        self._config = config
        self._file = file
        # This will contain the (lazy) imported gpg module:
        self._gpg = None

    def get_pass(self):
        contents = self._decrypt_file()
        if not contents:
            Logger.error(f"ERROR: Unable to decrypt file: {self._file}")
            return None
        lines = contents.decode().split('\n')
        if lines:
            # per 'pass' specification
            return(lines[0])
        return None

    def _decrypt_file(self):
        text = None
        if not self._gpg:
            self._gpg = importlib.import_module('gpg')
        with open(self._file, 'rb') as f:
            try:
                text, result, result = self._gpg.Context().decrypt(f)
            except self._gpg.errors.GPGMEError as e:
                Logger.error(f"ERROR: unable to decrypt file: {self._file}")
                Logger.error(e)
        return text
