# Copyright(c) 2020-2021 by craftyguy "Clayton Craft" <clayton@craftyguy.net>
# Distributed under GPLv3+ (see COPYING) WITHOUT ANY WARRANTY.
import gi
from .passlist import PassList
gi.require_version('Gtk', '3.0')
gi.require_version('Handy', '1')
from gi.repository import Gtk, Gio, GLib
from gi.repository import Handy

APP_NAME = 'caerbannog'
UI_PATH = f"/net/craftyguy/{APP_NAME}/ui/"


@Gtk.Template(resource_path=UI_PATH+'window.ui')
class MainWindow(Handy.ApplicationWindow):
    __gtype_name__ = 'MainWindow'
    mainbox: Gtk.Box = Gtk.Template.Child()
    headerbar: Handy.HeaderBar = Gtk.Template.Child()
    backbutton: Gtk.Button = Gtk.Template.Child()
    searchentry: Gtk.SearchEntry = Gtk.Template.Child()
    pass_store_list_scroll_window: Gtk.ScrolledWindow = Gtk.Template.Child()

    def __init__(self, *args, **kwargs):
        # process (and remove..) any args not supported by
        # Gtk.ApplicationWindow
        if 'config' in kwargs:
            self._config = kwargs.pop('config')
        else:
            raise RuntimeError("ERROR: config not provided")
        super().__init__(**kwargs)
        # self.connect("delete-event", self.on_delete_event)

        # instantiate here so config param can be passed to it
        self.pass_store_list = PassList(self._config)
        self.pass_store_list_scroll_window.add(self.pass_store_list)

        self.pass_store_list.get_keys_from_store()
        # set keyboard focus to search entry by default
        self.searchentry.grab_focus_without_selecting()

        # app starts at root of password store, so hide back button
        self.backbutton.hide()

        self.headerbar.set_title(APP_NAME)

        self.pass_store_list.update_view()

        # delayed search submission
        self._search_text = ''
        self._search_timer = None

        self._connect_signals()

        # Make everything visible (All is hidden by default in GTK 3)
        self.show_all()

    def _connect_signals(self):
        self.searchentry.connect("search-changed", self.on_search_changed)
        self.pass_store_list.connect("pass-level-changed",
                                     self.on_pass_level_changed)
        self.backbutton.connect("clicked", self.on_back_btn_clicked)

    def reload(self):
        self.pass_store_list.get_keys_from_store(force=True)
        self.show_all()

    def __add_action(self, action_name, callback):
        action = Gio.SimpleAction.new(action_name, None)
        action.connect("activate", callback)
        self.add_action(action)

    def on_back_btn_clicked(self, obj):
        self.pass_store_list.update_view(go_back=True)
        self._update_headerbar(self.pass_store_list.current_tree_level)

    def on_pass_level_changed(self, obj, title):
        self._update_headerbar(title)

    def _update_headerbar(self, title):
        if title:
            self.headerbar.set_title(title)
            self.backbutton.show()
        else:
            self.headerbar.set_title(APP_NAME)
            self.backbutton.hide()

    def on_search_changed(self, search_entry):
        self._search_text = search_entry.get_text()
        # use a timer to control when the search filtering is triggered
        if self._search_timer:
            # stop existing timer
            GLib.source_remove(self._search_timer)
        self._search_timer = GLib.timeout_add(666, self._search_cb)

    def _search_cb(self):
        """ Callback for searching, meant to be triggered by a timer """
        self.pass_store_list.update_filter(self._search_text)
        self._search_text = ''
        self._search_timer = None
        # disable the timer
        return False

    def on_leaflet_folded(self, *args):
        return
